import logging
from typing import Optional

import requests
from greynoise import GreyNoise


class ThreatData:
    __OPSWAT_KEY = None
    __OPSWAT_URL = "https://api.metadefender.com/v4/ip/"
    __GREYNOISE_KEY = None

    @staticmethod
    def set_opswat_key(opswat_key: str):
        ThreatData.__OPSWAT_KEY = opswat_key

    @staticmethod
    def set_greynoise_key(greynoise_key: str):
        ThreatData.__GREYNOISE_KEY = greynoise_key

    @staticmethod
    def fetch_country(ip: str):
        url = ThreatData.__OPSWAT_URL + ip
        headers = dict(apikey=ThreatData.__OPSWAT_KEY)
        res = requests.get(url=url, headers=headers)
        if res.status_code == 200:
            res = res.json()
            geo_info = res.get('geo_info', {})
            country = geo_info.get('country', {})
            country = country.get('name')
        else:
            country = None
        return country

    @staticmethod
    def fetch_OPSWAT_summary(ip: str) -> (list, list):
        tags, scans = list(), list()
        url = ThreatData.__OPSWAT_URL + ip
        headers = {'apikey': ThreatData.__OPSWAT_KEY}
        res = requests.get(url=url, headers=headers)
        if res.status_code == 200:
            res = res.json()
            res = res.get('lookup_results', {})
            sources = res.get('sources', [])
            for source in sources:
                scans.append(source['status'])
                if source['status'] not in [0, 5]:
                    tags.append(source['assessment'])
        else:
            logging.warning('Failed to fetch OPSWAT summary for ip {}'.format(ip))
        return tags, scans

    @staticmethod
    def fetch_TALOS_summary(ip: str) -> str:
        url = 'https://talosintelligence.com/sb_api/query_lookup'
        headers = {
            'referer': 'https://talosintelligence.com/reputation_center/lookup?search=%s' % ip,
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36'
        }
        params = {
            'query': '/api/v2/details/ip/',
            'query_entry': ip
        }
        res = requests.get(url=url, headers=headers, params=params)
        if res.status_code == 201:
            res_json = res.json()
            email_rep: str = res_json.get('email_score_name', 'Missing')
            web_rep: str = res_json.get('web_score_name', 'Missing')
            summary = f"Email Reputation - {email_rep.upper()}, Web Reputation - {web_rep.upper()}"
        else:
            logging.warning('Failed to fetch Talos summary for ip {}'.format(ip))
            summary = None
        return summary

    @staticmethod
    def fetch_GREYNOISE_summary(ip: str) -> (Optional[str], Optional[str], Optional[list]):
        gn = GreyNoise(api_key=ThreatData.__GREYNOISE_KEY, timeout=30)
        gn_ip_data = gn.ip(ip)
        organization = gn_ip_data.get('metadata', {}).get('organization')
        classification = gn_ip_data['classification'].upper() if 'classification' in gn_ip_data \
                                                                 and gn_ip_data['classification'] != 'unknown' else None
        tags = gn_ip_data.get('tags')
        return classification, organization, tags



from BlacklistReport.BlacklistReport import generate_blacklist_report
from BlacklistReport.BlacklistReport import BlacklistReport
from BlacklistReport.BlacklistEntry import BlacklistEntry

import logging


logging.basicConfig(level=logging.WARNING,
                    format='%(name)s - %(levelname)s - %(message)s')



import logging
from jsonschema import validate, ValidationError, SchemaError


INPUTTED_REPORT_SCHEMA = {
    "type": "object",
    "properties": {
        "incidents": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {"type": "string"},
                    "source": {
                        "type": "object",
                        "properties": {
                            "ip": {"type": "string"},
                            "domain": {"type": "string"},
                            "country": {"type": "string"},
                            "opswat_scans": {
                                "type": "array",
                                "items": {"type": "string"}
                            },
                            "opswat_tags": {
                                "type": "array",
                                "items": {"type": "string"}
                            },
                            "gn_classification": {"type": "string"},
                            "gn_organization": {"type": "string"},
                            "gn_tags": {
                                "type": "array",
                                "items": {"type": "string"}
                            }
                        }
                    },
                    "destination": {
                        "type": "object",
                        "properties": {
                            "ip": {"type": "string"},
                            "domain": {"type": "string"},
                            "country": {"type": "string"},
                            "opswat_scans": {
                                "type": "array",
                                "items": {"type": "string"}
                            },
                            "opswat_tags": {
                                "type": "array",
                                "items": {"type": "string"}
                            },
                            "gn_classification": {"type": "string"},
                            "gn_organization": {"type": "string"},
                            "gn_tags": {
                                "type": "array",
                                "items": {"type": "string"}
                            }
                        }
                    }
                }
            }
        }
    }
}

GC_CSV_SCHEMA = {
    "type": "object",
    "properties": {
        "id": {"type": "string"},
        "severity": {"enum": ["Low", "Medium", "High"]},
        "start_time (utc)": {"type": "string"},
        "incident_type": {"enum": ["Incident", "Deception", "Network Scan", "Reveal", "Experimental"]},
        "affected_assets": {"type": "string"},
        "tags": {"type": "string"},
        "incident_group": {"type": "string"},
        "exporting_error": {"type": "string"}
    },
    "required": ["id", "affected_assets"]
}

CENTRA_SCHEMA = {
    "type": "object",
    "properties": {
        "username": {"type": "string"},
        "dev_host": {"type": "string"},
        "password": {"type": "string"}
    },
    "required": ["username", "dev_host", "password"]
}


def is_valid_format(entity, schema) -> bool:
    valid = False
    try:
        validate(entity, schema)
        valid = True
    except (ValidationError, SchemaError):
        logging.error(f"Object validation failed.", exc_info=True)
    finally:
        return valid



from typing import List, Union
from ipaddress import ip_address, IPv4Address, IPv6Address

from BlacklistReport.__threat_data import ThreatData


class BlacklistEntry:
    def __init__(self, source: dict, destination: dict):
        self.source = self.IpEntry(**source)
        self.__destinations: List = [self.IpEntry(**destination)]

    def add_destination(self, destination: dict):
        for entry in self.__destinations:
            if entry.ip == destination['ip']:
                return
        self.__destinations.append(self.IpEntry(**destination))

    def to_dict(self):
        return dict(source=self.source.to_dict(), destinations=[d.to_dict() for d in self.destinations])

    @property
    def destinations(self):
        return self.__destinations.copy()

    class IpEntry:

        def __init__(
                self,
                ip: str,
                ports: List[int] = None,
                domain: str = None,
                country: str = None,
                opswat_scans: List[str] = None,
                opswat_tags: List[str] = None,
                gn_classification: str = None,
                gn_organization: str = None,
                gn_tags: List[str] = None
        ):
            try:
                self.__ip: Union[IPv4Address, IPv6Address] = ip_address(address=ip)
            except ValueError:
                self.__ip = ip
            finally:
                self.__ports = ports or list()
                self.domain = domain
                self.__country = country
                self.opswat_scans = opswat_scans
                self.opswat_tags = opswat_tags
                self.gn_classification = gn_classification
                self.gn_organization = gn_organization
                self.gn_tags = gn_tags
                if type(self.__ip) in [IPv4Address, IPv6Address] and self.__ip.is_global:
                    if not self.__country:
                        self.__country = ThreatData.fetch_country(ip)
                    if not any([self.opswat_tags, self.opswat_scans]):
                        self.opswat_tags, self.opswat_scans = ThreatData.fetch_OPSWAT_summary(ip)
                    if not any([self.gn_classification, self.gn_organization, self.gn_tags]):
                        self.gn_classification, self.gn_organization, self.gn_tags = ThreatData.fetch_GREYNOISE_summary(ip)
                else:
                    self.__country = 'Internal'

        def to_dict(self) -> dict:
            return dict(
                ip=self.ip,
                domain=self.domain,
                ports=self.ports,
                origin_country=self.country,
                opswat=self.opswat,
                greynoise=self.greynoise
            )

        @property
        def ip(self) -> str:
            return str(self.__ip)

        @property
        def ports(self) -> str:
            return ', '.join(list(set([str(port) for port in self.__ports])))

        @property
        def country(self):
            return self.__country or 'Unavailable'

        @property
        def opswat(self):
            summary = None
            detected = [scan for scan in self.opswat_scans if str(scan) not in ("0", "5")]
            if any([self.opswat_scans, self.opswat_tags]):
                summary = f"{len(detected)}/{len(self.opswat_scans)} Score"
                if len(self.opswat_tags) > 0:
                    summary += " - " + (', '.join(self.opswat_tags)).title()
            return summary or 'Unavailable'

        @property
        def greynoise(self):
            summary_data = list()
            if self.gn_organization:
                summary_data.append(f"Organization - {self.gn_organization}")
            if self.gn_classification:
                summary_data.append(f"Classification - {self.gn_classification.upper()}")
            if self.gn_tags:
                summary_data.append(f"Tags - {self.gn_tags}")
            return ', '.join(summary_data) or 'Unavailable'



import csv
import logging
from datetime import timedelta
from typing import TypeVar, List, Dict, Optional

import inflect
from GuardiPy import Centra, CentraEntity, CentraApiPayload
from GuardiPy.CentraObject import Incident, LabelMinimal

from BlacklistReport.BlacklistEntry import BlacklistEntry
from BlacklistReport.__threat_data import ThreatData
from BlacklistReport.schema import GC_CSV_SCHEMA, CENTRA_SCHEMA, is_valid_format

AnyCentraEntity = TypeVar("AnyCentraEntity", bound=CentraEntity)


def deduplicate_incidents(incidents: list) -> list:
    unique_incidents = dict()
    for i in incidents:
        if f"{i['source']['ip']}-{i['destination']['ip']}" not in unique_incidents:
            unique_incidents[f"{i['source']['ip']}-{i['destination']['ip']}"] = i
    return list(unique_incidents.values())


def generate_blacklist_report(gc_host: Dict, hours: int, opswat_key: str, greynoise_key: str,
                              gc_customer_name: str, severity: List[str] = None, incidents: List[Dict] = None):
    if is_valid_format(gc_host, CENTRA_SCHEMA):

        gc = Centra(username=gc_host['username'],
                    hostname=gc_host['dev_host'],
                    password=gc_host['password'])

        converter = inflect.engine()

        ThreatData.set_opswat_key(opswat_key=opswat_key)
        ThreatData.set_greynoise_key(greynoise_key=greynoise_key)

        blr = BlacklistReport(gc=gc, gc_customer_name=gc_customer_name, incidents=incidents)
        incident_count = blr.fetch_incidents(hours, severity=severity)
        incident_count_str = converter.number_to_words(incident_count)
        blr.build_report()

        source_ip_count = len(blr.report.keys())
        source_ip_count_str = converter.number_to_words(source_ip_count)

        return dict(
            hours=hours,
            incident_count=incident_count,
            incident_count_str=incident_count_str,
            report=blr.build_report_str(incident_count_str, incident_count, source_ip_count_str, source_ip_count)
        )

    else:
        logging.error("Failed to instantiate Centra instance.", exc_info=True)


class BlacklistReport:
    def __init__(self, gc: Centra, gc_customer_name: str, incidents: List[Dict] = None):
        self.gc: Centra = gc
        self.incidents: List[Dict] = deduplicate_incidents(incidents) if incidents else list()
        self.report: Dict = dict()
        self.hours: int = -1
        self.shared_instance, self.customer_label = self.fetch_customer_label(gc_customer_name)

    def fetch_customer_label(self, gc_customer_name: str) -> (bool, Optional[str]):
        shared_instance = False
        customer_label = None
        query: CentraApiPayload = LabelMinimal.list(
            assets='on,off',
            find_matches=True,
            dynamic_criteria_limit=500,
            key='Customers'
        )
        res = self.gc.execute(query)
        if res:
            shared_instance = True
            for i in res:
                if i.value == gc_customer_name:
                    customer_label = i.id
        return shared_instance, customer_label

    def fetch_incidents(self, hours: int = 24, severity: List[str] = None) -> int:
        self.hours = abs(hours)
        if not self.incidents and (not self.shared_instance or self.customer_label):
            query = Incident.list(
                from_time=timedelta(hours=-self.hours),
                incident_type='Reveal',
                tags_include='Reputation',
                severity=severity or ["Medium", "High"],
                prefixed_filter='bad_reputation'
            )
            if self.customer_label:
                query.params['any_side'] = self.customer_label
            result = self.gc.export_to_csv(query).splitlines()
            csv_reader = csv.DictReader(result, delimiter=',')
            self.incidents = [_parse_reported_fields(row) for row in csv_reader]
        for incident in self.incidents:
            self._enrich_incident_data(incident)
        logging.debug("Number of blacklisted IP incidents: %d", len(self.incidents))
        return len(self.incidents)

    def build_report(self):
        for incident in self.incidents:
            source = incident['source']
            destination = incident['destination']
            if source['ip'] in self.report:
                self.report[source['ip']].add_destination(destination=destination)
            else:
                self.report[source['ip']] = BlacklistEntry(source=source, destination=destination)
        return self.report.copy()

    def _enrich_incident_data(self, incident: dict) -> None:
        if not (all([k in incident['source'] for k in ['ip']])
                and all([k in incident['destination'] for k in ['ip', 'ports']])):
            info: Incident = self.gc.execute(Incident.get_all_info(id=incident['id']))
            if info:
                source = incident['source']
                destination = incident['destination']
                if not source.get('ip'):
                    source['ip'] = info.source_asset.ip
                if not source.get('domain'):
                    source['domain'] = info.source_asset.vm.name if info.source_asset.vm else None
                if not source.get('country'):
                    source['country'] = info.source_asset.country
                if not destination.get('ip'):
                    destination['ip'] = info.destination_asset.ip
                if not destination.get('domain'):
                    destination['domain'] = info.destination_asset.vm.name \
                        if info.destination_asset.vm else None
                if not destination.get('ports'):
                    destination['ports'] = list(set([e.destination_port for e in info.events
                                                     if e.destination_port is not None and e.destination_port >= 0]))

    def build_report_str(self, incident_count_str: str, incident_count: int, source_ip_count_str: str,
                         source_ip_count: int):
        report_summary = self._plaintext_header(incident_count_str, incident_count, source_ip_count_str,
                                                source_ip_count)
        report_summary += ''.join([_plaintext_entry(ip, entry) for ip, entry in self.report.items()])
        return report_summary

    def _plaintext_header(self, incident_count_str, incident_count, source_ip_count_str, source_ip_count) -> str:
        return f"In the past {self.hours} hours, we have seen {incident_count_str} ({incident_count}) " \
               f"reputation incidents from {source_ip_count_str} ({source_ip_count}) sources.\n\n"


def _parse_reported_fields(entity: Dict) -> Dict:
    parsed_entity = dict()
    if is_valid_format(entity, GC_CSV_SCHEMA):
        affected_assets = [asset.strip() for asset in entity.get('affected_assets', '').split('),')]
        source_ip = affected_assets[0].split(' ')[0]
        destination_ip = affected_assets[1].split(' ')[0]
        if "\\" in destination_ip:
            destination_ip = destination_ip.split("\\")[1]
        parsed_entity = dict(id=entity['id'], source=dict(ip=source_ip), destination=dict(ip=destination_ip))
    return parsed_entity


def _plaintext_entry(ip, ip_entry: BlacklistEntry) -> str:
    text = ""
    text += f"Source: {ip} ({ip_entry.source.country})\n"
    if ip_entry.source.country != 'Internal':
        text += f"\tOPSWAT: {ip_entry.source.opswat}\n"
        text += f"\tGreyNoise: {ip_entry.source.greynoise}\n"
    text += f"Destination(s):\n"
    for destination in ip_entry.destinations:
        text += f"\tDestination: {destination.ip} ({destination.country})\n"
        text += f"\tDestination Port(s): {destination.ports}"
        if destination.country != 'Internal':
            text += f"\n\tOPSWAT: {destination.opswat}"
            text += f"\n\tGreyNoise: {ip_entry.source.greynoise}\n"
        else:
            text += "\n"
        text += "\n"
    text += "\n"
    return text



import setuptools

with open("README.MD", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="BlacklistReport",
    version="0.4.1",
    author="dzellmer",
    description="Bad Reputation (Blacklisted IP) Incident Reporting.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://git.vcjames.io/soc/dailyblacklistreporting",
    packages=setuptools.find_packages(),
    install_requires=required,
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
